# Copyright (c) 2022 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import pytest
import numpy as np


def test_lagrange_custom_triangle_degree1():
    """Test that Lagrange element created as a custom element agrees with built-in Lagrange."""

    wcoeffs = np.eye(3)
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]])],
         [z, z, z], [z], []]
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
         [z, z, z], [z], []]

    lagrange = basix.create_element(
        basix.ElementFamily.P, basix.CellType.triangle, 1)

    element = basix.create_custom_element(
        basix.CellType.triangle, [], wcoeffs,
        x, M, basix.MapType.identity, False, 1, 1)

    points = basix.create_lattice(basix.CellType.triangle, 5, basix.LatticeType.equispaced, True)
    assert np.allclose(lagrange.tabulate(1, points), element.tabulate(1, points))
    assert np.allclose(lagrange.base_transformations(), element.base_transformations())


def test_lagrange_custom_triangle_degree1_l2piola():
    """Test a custom element with a L2 Piola map."""

    wcoeffs = np.eye(3)
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]])],
         [z, z, z], [z], []]
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
         [z, z, z], [z], []]

    lagrange = basix.create_element(
        basix.ElementFamily.P, basix.CellType.triangle, 1)

    element = basix.create_custom_element(
        basix.CellType.triangle, [], wcoeffs,
        x, M, basix.MapType.L2Piola, False, 1, 1)

    points = basix.create_lattice(basix.CellType.triangle, 5, basix.LatticeType.equispaced, True)
    assert np.allclose(lagrange.tabulate(1, points), element.tabulate(1, points))
    assert np.allclose(lagrange.base_transformations(), element.base_transformations())


def test_lagrange_custom_triangle_degree4():
    """Test that Lagrange element created as a custom element agrees with built-in Lagrange."""

    wcoeffs = np.eye(15)
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]])],
         [np.array([[.75, .25], [.5, .5], [.25, .75]]), np.array([[0., .25], [0., .5], [0., .75]]),
          np.array([[.25, 0.], [.5, 0.], [.75, 0.]])],
         [np.array([[.25, .25], [.5, .25], [.25, .5]])], []]
    id = np.array([[[1., 0., 0.]], [[0., 1., 0.]], [[0., 0., 1.]]])
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
         [id, id, id], [id], []]

    lagrange = basix.create_element(
        basix.ElementFamily.P, basix.CellType.triangle, 4, basix.LagrangeVariant.equispaced)

    element = basix.create_custom_element(
        basix.CellType.triangle, [], wcoeffs,
        x, M, basix.MapType.identity, False, 4, 4)

    points = basix.create_lattice(basix.CellType.triangle, 5, basix.LatticeType.equispaced, True)
    assert np.allclose(lagrange.tabulate(1, points), element.tabulate(1, points))
    assert np.allclose(lagrange.base_transformations(), element.base_transformations())


def test_lagrange_custom_quadrilateral_degree1():
    """Test that Lagrange element created as a custom element agrees with built-in Lagrange."""

    wcoeffs = np.eye(4)
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]]), np.array([[1., 1.]])],
         [z, z, z, z], [z], []]
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
         [z, z, z, z], [z], []]

    lagrange = basix.create_element(
        basix.ElementFamily.P, basix.CellType.quadrilateral, 1)

    element = basix.create_custom_element(
        basix.CellType.quadrilateral, [], wcoeffs,
        x, M, basix.MapType.identity, False, 1, 1)

    points = basix.create_lattice(basix.CellType.quadrilateral, 5, basix.LatticeType.equispaced, True)
    assert np.allclose(lagrange.tabulate(1, points), element.tabulate(1, points))
    assert np.allclose(lagrange.base_transformations(), element.base_transformations())


def test_raviart_thomas_triangle_degree1():
    """Test that Raviart-Thomas element created as a custom element agrees with built-in Raviart-Thomas."""

    wcoeffs = np.zeros((3, 6))
    wcoeffs[0, 0] = 1
    wcoeffs[1, 3] = 1

    pts, wts = basix.make_quadrature(basix.CellType.triangle, 2)
    poly = basix.tabulate_polynomials(basix.PolynomialType.legendre, basix.CellType.triangle, 1, pts)
    for i in range(3):
        wcoeffs[2, i] = sum(pts[:, 0] * poly[:, i] * wts)
        wcoeffs[2, 3 + i] = sum(pts[:, 1] * poly[:, i] * wts)

    pts, wts = basix.make_quadrature(basix.CellType.interval, 1)

    x = [[], [], [], []]
    for _ in range(3):
        x[0].append(np.zeros((0, 2)))
    x[1].append(np.array([[1 - p[0], p[0]] for p in pts]))
    x[1].append(np.array([[0, p[0]] for p in pts]))
    x[1].append(np.array([[p[0], 0] for p in pts]))
    x[2].append(np.zeros((0, 2)))

    M = [[], [], [], []]
    for _ in range(3):
        M[0].append(np.zeros((0, 2, 0)))
    for normal in [[-1, -1], [-1, 0], [0, 1]]:
        M[1].append(np.array([[normal[0] * wts, normal[1] * wts]]))
    M[2].append(np.zeros((0, 2, 0)))

    element = basix.create_custom_element(
        basix.CellType.triangle, [2], wcoeffs, x, M, basix.MapType.contravariantPiola, False, 0, 1)

    rt = basix.create_element(
        basix.ElementFamily.RT, basix.CellType.triangle, 1)

    points = basix.create_lattice(basix.CellType.triangle, 5, basix.LatticeType.equispaced, True)
    assert np.allclose(rt.tabulate(1, points), element.tabulate(1, points))
    assert np.allclose(rt.base_transformations(), element.base_transformations())


def create_lagrange1_quad(cell_type=basix.CellType.quadrilateral, degree=1, wcoeffs=None, x=None, M=None,
                          value_shape=None, discontinuous=False):
    """Attempt to create a Lagrange 1 element on a quad."""
    if wcoeffs is None:
        wcoeffs = np.eye(4)
    if x is None:
        z = np.zeros((0, 2))
        x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]]), np.array([[1., 1.]])],
             [z, z, z, z], [z], []]
    if M is None:
        z = np.zeros((0, 1, 0))
        M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
             [z, z, z, z], [z], []]
    if value_shape is None:
        value_shape = []
    basix.create_custom_element(
        cell_type, value_shape, wcoeffs, x, M, basix.MapType.identity, discontinuous, 1, degree)


def test_create_lagrange1_quad():
    """Test that the above function works if no inputs are modified."""
    create_lagrange1_quad()


def assert_failure(**kwargs):
    """Assert that the correct RuntimeError is thrown."""
    try:
        create_lagrange1_quad(**kwargs)
    except RuntimeError as e:
        if len(e.args) == 0:
            raise e
        if "dgesv" in e.args[0]:
            raise e
        return
    with pytest.raises(RuntimeError):
        pass


def test_wcoeffs_wrong_shape():
    """Test that a runtime error is thrown when wcoeffs is the wrong shape."""
    assert_failure(wcoeffs=np.eye(3))


def test_wcoeffs_too_few_cols():
    """Test that a runtime error is thrown when wcoeffs has too few columns."""
    assert_failure(wcoeffs=np.array([[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., 0.]]))


def test_wcoeffs_too_few_rows():
    """Test that a runtime error is thrown when wcoeffs has too few rows."""
    assert_failure(wcoeffs=np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.], [1., 0., 1.]]))


def test_wcoeffs_zero_row():
    """Test that a runtime error is thrown when wcoeffs has a row of zeros."""
    assert_failure(wcoeffs=np.array([[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., 0.], [0., 0., 0., 0.]]))


def test_wcoeffs_equal_rows():
    """Test that a runtime error is thrown when wcoeffs has two equal rows."""
    assert_failure(wcoeffs=np.array([[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., 1.], [0., 0., 1., 1.]]))


def test_x_wrong_tdim():
    """Test that a runtime error is thrown when a point in x has the wrong tdim."""
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]]), np.array([[1., 1.]])],
         [z, z, z, z], [z], []]
    assert_failure(x=x)


def test_x_wrong_ndims():
    """Test that a runtime error is thrown when x has the wrong shape."""
    z = np.zeros((0, 2))
    x = [[np.array([0., 0.]), np.array([[1., 0.]]), np.array([[0., 1.]]), np.array([[1., 1.]])],
         [z, z, z, z], [z], []]
    assert_failure(x=x)


def test_x_too_many_points():
    """Test that a runtime error is thrown when x has too many points."""
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]]), np.array([[1., 1.]])],
         [z, z, z, z], [np.array([[.5, .5]])], []]
    assert_failure(x=x)


def test_x_point_tdim_too_high():
    """Test that a runtime error is thrown when x has a point in an entity with a too high dimension."""
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]]), z],
         [z, z, z, z], [z], [np.array([[1., 1.]])]]
    assert_failure(x=x)


def test_x_wrong_entity_count():
    """Test that a runtime error is thrown when x has the wrong number of edges."""
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]]), np.array([[1., 1.]])],
         [z, z, z, z, z], [z], []]
    assert_failure(x=x)


def test_M_wrong_value_size():
    """Test that a runtime error is thrown when M has the wrong shape."""
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.], [1.]]]), np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
         [z, z, z, z], [z], []]
    assert_failure(M=M)


def test_M_wrong_ndims():
    """Test that a runtime error is thrown when M has the wrong shape."""
    z = np.zeros((0, 1, 0))
    M = [[np.array([[1.]]), np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
         [z, z, z, z], [z], []]
    assert_failure(M=M)


def test_M_too_many_points():
    """Test that a runtime error is thrown when M is the wrong shape."""
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])],
         [z, z, z, z], [np.array([[[1.]]])], []]
    assert_failure(M=M)


def test_M_wrong_entities():
    """Test that a runtime error is thrown when the shape of M does not match x."""
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]], [[1.]]]), z],
         [z, z, z, z], [z], []]
    assert_failure(M=M)


def test_M_value_size():
    """Test that a runtime error is thrown when M is the wrong shape."""
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.], [1.]]]), np.array([[[1.]]])],
         [z, z, z, z], [z], []]
    assert_failure(M=M)


def test_M_zero_row():
    """Test that a runtime error is thrown when M has a zero row."""
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[0.]]])],
         [z, z, z, z], [z], []]
    assert_failure(M=M)


def test_wrong_value_shape():
    """Test that a runtime error is thrown when value shape is wrong."""
    assert_failure(value_shape=[2])


def test_wrong_cell_type():
    """Test that a runtime error is thrown when cell type is wrong."""
    assert_failure(cell_type=basix.CellType.hexahedron)


def test_wrong_degree():
    """Test that a runtime error is thrown when degree is wrong."""
    assert_failure(degree=0)


def test_wrong_discontinuous():
    """Test that a runtime error is thrown when discontinuous is wrong."""
    assert_failure(discontinuous=True)



# Copyright (c) 2022 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import pytest

cells = [
    basix.CellType.point,
    basix.CellType.interval,
    basix.CellType.triangle,
    basix.CellType.quadrilateral,
    basix.CellType.tetrahedron,
    basix.CellType.hexahedron,
    basix.CellType.prism,
    basix.CellType.pyramid,
]

elements = [
    basix.ElementFamily.P,
    basix.ElementFamily.RT,
    basix.ElementFamily.BDM,
    basix.ElementFamily.N1E,
    basix.ElementFamily.N2E,
    basix.ElementFamily.Regge,
    basix.ElementFamily.HHJ,
    basix.ElementFamily.bubble,
    basix.ElementFamily.serendipity,
    basix.ElementFamily.DPC,
    basix.ElementFamily.CR,
    basix.ElementFamily.custom,
]

variants = [
    [basix.LagrangeVariant.gll_isaac],
    [basix.LagrangeVariant.gll_warped],
    [basix.LagrangeVariant.legendre],
    [basix.DPCVariant.diagonal_gll],
    [basix.DPCVariant.legendre],
    [basix.LagrangeVariant.legendre, basix.DPCVariant.legendre],
]


def test_all_cells_included():
    all_cells = set()
    for c in dir(basix.CellType):
        if not c.startswith("_") and c not in ["name", "value"]:
            all_cells.add(getattr(basix.CellType, c))

    assert all_cells == set(cells)


def test_all_elements_included():
    all_elements = set()
    for c in dir(basix.ElementFamily):
        if not c.startswith("_") and c not in ["name", "value"]:
            all_elements.add(getattr(basix.ElementFamily, c))

    assert all_elements == set(elements)


@pytest.mark.parametrize("cell", cells)
@pytest.mark.parametrize("degree", range(-1, 5))
@pytest.mark.parametrize("family", elements)
@pytest.mark.parametrize("variant", variants)
def test_create_element(cell, degree, family, variant):
    """Check that either the element is created or a RuntimeError is thrown."""
    try:
        element = basix.create_element(family, cell, degree, *variant)
        assert element.degree == degree
    except RuntimeError as e:
        # Don't allow cryptic "dgesv failed" messages
        if len(e.args) == 0 or "dgesv" in e.args[0]:
            raise e

    try:
        element = basix.create_element(family, cell, degree, *variant, True)
        assert element.degree == degree
    except RuntimeError as e:
        if len(e.args) == 0 or "dgesv" in e.args[0]:
            raise e



# =============================
# Variants of Lagrange elements
# =============================
#
# When creating high degree function spaces, it is important to define the
# degrees of freedom (DOFs) of a space in a way that leads to well behaved
# basis functions. For example, if a Lagrange space is created using equally
# spaced points, then the basis functions with exhibit Runge's phenomenon
# and large peaks will be observed near the edges of the cell.
#
# When a finite element is defined using points evaluation DOFs, we can use
# the Lebesgue constant to indicate how well behaved a set of basis functions
# will be. The Lebesgue constant is defined by
#
# .. math::
#    \Lambda = \max_{x\in R}\left(\sum_i\left|\phi_i(x)\right|\right),
#
# where :math:`R` is the reference cell and :math:`\phi_0` to :math:`\phi_{n-1}`
# are the basis functions of the finite element space. A smaller value of
# :math:`\Lambda` indicates a better set of basis functions.
#
# In this demo, we look at how the Lebesgue constant can be approximated using
# Basix, and how variants of Lagrange elements that have lower Lebesgue constants
# can be created.
#
# We begin by importing Basix and Numpy.

import basix
import numpy as np
from basix import ElementFamily, CellType, LagrangeVariant, LatticeType

# In this demo, we consider Lagrange elements defined on a triangle. We start
# by creating a degree 15 Lagrange element that uses equally spaced points.
# This element will exhibit Runge's phenomenon, so we expect a large Lebesgue
# constant.

lagrange = basix.create_element(
    ElementFamily.P, CellType.triangle, 15, LagrangeVariant.equispaced)

# To estimate the Lebesgue constant, we create a lattice of points on the
# triangle and compute
#
# .. math::
#    \Lambda \approx \max_{x\in L}\left(\sum_i\left|\phi_i(x)\right|\right),
#
# where :math:`L` is the set of points in our lattice. As :math:`L` is a
# subset of :math:`R`, the values we compute will be lower bounds of the true
# Lebesgue constants.
#
# The function `create_lattice` takes four inputs: the cell type, the number
# of points in each direction, the lattice type (in this example, we use an
# equally spaced lattice), and a bool indicating whether or not points on the
# boundary should be included. We tabulate our element at the points in the
# lattice then use Numpy to compute the max of the sum.
#
# As expected, the value is large.

points = basix.create_lattice(
    CellType.triangle, 50, LatticeType.equispaced, True)
tab = lagrange.tabulate(0, points)[0]
print(max(np.sum(np.abs(tab), axis=0)))

# A Lagrange element with a lower Lebesgue constant can be created by placing
# the DOFs at Gauss-Lobatto-Legendre (GLL) points. Passing
# `LagrangeVariant.gll_warped` into `create_element` will make an element that
# places its DOF points at warped GLL points on the triangle, as described in
# `Nodal Discontinuous Galerkin Methods (Hesthaven, Warburton, 2008, pp
# 175-180) <https://doi.org/10.1007/978-0-387-72067-8>`_.
#
# The Lebesgue constant for this variant of the element is much smaller than
# for the equally spaced element.

gll = basix.create_element(
    ElementFamily.P, CellType.triangle, 15, LagrangeVariant.gll_warped)
print(max(np.sum(np.abs(gll.tabulate(0, points)[0]), axis=0)))

# An even lower Lebesgue constant can be obtained by placing the DOF points
# at GLL points mapped onto a triangle following the method proposed in
# `Recursive, Parameter-Free, Explicitly Defined Interpolation Nodes for
# Simplices (Isaac, 2020) <https://doi.org/10.1137/20M1321802>`_.

gll2 = basix.create_element(
    ElementFamily.P, CellType.triangle, 15, LagrangeVariant.gll_isaac)
print(max(np.sum(np.abs(gll2.tabulate(0, points)[0]), axis=0)))



"""Functions to manipulate variant types."""

from ._basixcpp import LagrangeVariant as _LV
from ._basixcpp import DPCVariant as _DV


def string_to_lagrange_variant(variant: str) -> _LV:
    """Convert a string to a Basix LagrangeVariant enum.

    Args:
        variant: The Lagrange variant as a string.

    Returns:
        The Lagrange variant
    """
    if variant.lower() == "gll":
        return _LV.gll_warped
    if variant.lower() == "chebyshev":
        return _LV.chebyshev_isaac
    if variant.lower() == "gl":
        return _LV.gl_isaac

    if not hasattr(_LV, variant.lower()):
        raise ValueError(f"Unknown variant: {variant}")
    return getattr(_LV, variant.lower())


def lagrange_variant_to_string(variant: _LV) -> str:
    """Convert a Basix LagrangeVariant enum to a string.

    Args:
        variant: The Lagrange variant

    Returns:
        The Lagrange variant as a string.
    """
    return variant.name


def string_to_dpc_variant(variant: str) -> _DV:
    """Convert a string to a Basix DPCVariant enum.

    Args:
        variant: The DPC variant as a string.

    Returns:
        The DPC variant
    """
    if not hasattr(_DV, variant.lower()):
        raise ValueError(f"Unknown variant: {variant}")
    return getattr(_DV, variant.lower())


def dpc_variant_to_string(variant: _DV) -> str:
    """Convert a Basix DPCVariant enum to a string.

    Args:
        variant: The DPC variant

    Returns:
        The DPC variant as a string.
    """
    return variant.name



# Copyright (c) 2020 Chris Richardson & Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import sympy
import basix
import numpy as np


def P_interval(n, x):
    r = []
    for i in range(n + 1):
        p = x ** i
        for j in r:
            p -= (p * j).integrate((x, 0, 1)) * j
        p /= sympy.sqrt((p * p).integrate((x, 0, 1)))
        r.append(p)
    return r


def test_symbolic_interval():
    n = 7
    nderiv = 7

    x = sympy.Symbol("x")
    wd = P_interval(n, x)

    cell = basix.CellType.interval
    pts0 = basix.create_lattice(cell, 10, basix.LatticeType.equispaced, True)
    wtab = basix._basixcpp.tabulate_polynomial_set(cell, n, nderiv, pts0)

    for k in range(nderiv + 1):
        wsym = np.zeros_like(wtab[k])
        for i in range(n + 1):
            for j, p in enumerate(pts0):
                wsym[j, i] = wd[i].subs(x, p[0])
            wd[i] = sympy.diff(wd[i], x)
        assert np.allclose(wtab[k], wsym)


def test_symbolic_quad():
    n = 5
    nderiv = 5

    idx = basix.index

    x = sympy.Symbol("x")
    y = sympy.Symbol("y")
    w = [wx * wy for wx in P_interval(n, x) for wy in P_interval(n, y)]

    m = (n + 1)**2
    cell = basix.CellType.quadrilateral
    pts0 = basix.create_lattice(cell, 2, basix.LatticeType.equispaced, True)
    wtab = basix._basixcpp.tabulate_polynomial_set(cell, n, nderiv, pts0)

    for kx in range(nderiv):
        for ky in range(0, nderiv - kx):

            wsym = np.zeros_like(wtab[0])
            for i in range(m):
                wd = sympy.diff(w[i], x, kx, y, ky)
                for j, p in enumerate(pts0):
                    wsym[j, i] = wd.subs([(x, p[0]), (y, p[1])])
            assert np.allclose(wtab[idx(kx, ky)], wsym)



# Copyright (c) 2021 Matthew Scroggs, Igor A. Baratta
# FEniCS Project
# SPDX-License-Identifier: MIT

import numpy as np
import basix
import pytest
from itertools import product
import numpy
from .utils import parametrize_over_elements


def tensor_product(*data):
    if len(data) == 1:
        return data[0]
    if len(data) > 2:
        return tensor_product(tensor_product(data[0], data[1]), *data[2:])

    a, b = data
    return np.outer(a, b).reshape(-1)


@parametrize_over_elements(4)
def test_tensor_product_factorisation(cell_type, degree, element_type, element_args):
    element = basix.create_element(element_type, cell_type, degree, *element_args)
    tdim = len(basix.topology(cell_type)) - 1

    # These elements should have a factorisation
    if cell_type in [
        basix.CellType.quadrilateral, basix.CellType.hexahedron
    ] and element_type in [
        basix.ElementFamily.P
    ] and basix.LagrangeVariant.equispaced in element_args:
        assert element.has_tensor_product_factorisation

    if not element.has_tensor_product_factorisation:
        with pytest.raises(RuntimeError):
            element.get_tensor_product_representation()
        pytest.skip()

    factors = element.get_tensor_product_representation()

    lattice = basix.create_lattice(cell_type, 1, basix.LatticeType.equispaced, True)
    tab1 = element.tabulate(1, lattice)

    for i, point in enumerate(lattice):
        for ds in product(range(2), repeat=tdim):
            if sum(ds) > 1:
                continue
            deriv = basix.index(*ds)
            values1 = tab1[deriv, i, :, :]

            values2 = np.empty((element.dim, 1))
            for fs, perm in factors:
                evals = [e.tabulate(d, p.reshape(1, -1))[d, 0, :, 0] for e, p, d in zip(fs, point, ds)]
                tab2 = tensor_product(*evals)
                for a, b in enumerate(perm):
                    if b != -1:
                        values2[b, 0] = tab2[a]
            assert np.allclose(values1, values2)


@pytest.mark.parametrize("degree", range(1, 9))
def test_tensor_product_factorisation_quadrilateral(degree):
    P = degree
    cell_type = basix.CellType.quadrilateral
    element = basix.create_element(basix.ElementFamily.P, cell_type,
                                   P, basix.LagrangeVariant.gll_warped)
    factors = element.get_tensor_product_representation()[0]

    # Quadrature degree
    Q = 2 * P + 2
    points, w = basix.make_quadrature(basix.QuadratureType.Default, cell_type, Q)
    data = element.tabulate(1, points)
    dphi_x = data[1, :, :, 0]
    dphi_y = data[2, :, :, 0]

    assert points.shape[0] == (P+2) * (P+2)

    # FIXME: This test assumes all factors formed by a single element
    perm = factors[1]
    element0 = factors[0][0]
    cell1d = element0.cell_type
    points, w = basix.make_quadrature(basix.QuadratureType.Default, cell1d, Q)
    data = element0.tabulate(1, points)
    phi0 = data[0, :, :, 0]
    dphi0 = data[1, :, :, 0]

    # number of dofs in each direction
    Nd = P + 1
    # number of quadrature points in each direction
    Nq = P + 2

    # Compute derivative of basis function in the x direction
    dphi_tensor = numpy.zeros([Nq, Nq, Nd, Nd])
    for q0 in range(Nq):
        for q1 in range(Nq):
            for i0 in range(Nd):
                for i1 in range(Nd):
                    dphi_tensor[q0, q1, i0, i1] = dphi0[q0, i0]*phi0[q1, i1]
    dphi_tensor = dphi_tensor.reshape([Nq*Nq, Nd*Nd])
    assert numpy.allclose(dphi_x[:, perm], dphi_tensor)

    # Compute derivative of basis function in the y direction
    dphi_tensor = numpy.zeros([Nq, Nq, Nd, Nd])
    for q0 in range(Nq):
        for q1 in range(Nq):
            for i0 in range(Nd):
                for i1 in range(Nd):
                    dphi_tensor[q0, q1, i0, i1] = phi0[q0, i0]*dphi0[q1, i1]

    dphi_tensor = dphi_tensor.reshape([Nq*Nq, Nd*Nd])
    assert numpy.allclose(dphi_y[:, perm], dphi_tensor)


@pytest.mark.parametrize("degree", range(1, 6))
def test_tensor_product_factorisation_hexahedron(degree):
    P = degree
    element = basix.create_element(basix.ElementFamily.P, basix.CellType.hexahedron,
                                   P, basix.LagrangeVariant.gll_warped)
    factors = element.get_tensor_product_representation()[0]

    # Quadrature degree
    Q = 2 * P + 2
    points, w = basix.make_quadrature(basix.QuadratureType.Default, basix.CellType.hexahedron, Q)
    data = element.tabulate(1, points)
    dphi_x = data[1, :, :, 0]
    dphi_y = data[2, :, :, 0]
    dphi_z = data[3, :, :, 0]

    assert points.shape[0] == (P+2) * (P+2) * (P+2)

    # FIXME: This test assumes all factors formed by a single element
    perm = factors[1]
    element0 = factors[0][0]
    cell1d = element0.cell_type
    points, w = basix.make_quadrature(basix.QuadratureType.Default, cell1d, Q)
    data = element0.tabulate(1, points)
    phi0 = data[0, :, :, 0]
    dphi0 = data[1, :, :, 0]

    # number of dofs in each direction
    Nd = P + 1
    # number of quadrature points in each direction
    Nq = P + 2

    # Compute derivative of basis function in the x direction
    dphi_tensor = numpy.zeros([Nq, Nq, Nq, Nd, Nd, Nd])
    for q0 in range(Nq):
        for q1 in range(Nq):
            for q2 in range(Nq):
                for i0 in range(Nd):
                    for i1 in range(Nd):
                        for i2 in range(Nd):
                            dphi_tensor[q0, q1, q2, i0, i1, i2] = dphi0[q0, i0]*phi0[q1, i1]*phi0[q2, i2]

    dphi_tensor = dphi_tensor.reshape([Nq*Nq*Nq, Nd*Nd*Nd])
    assert numpy.allclose(dphi_x[:, perm], dphi_tensor)

    # Compute derivative of basis function in the y direction
    dphi_tensor = numpy.zeros([Nq, Nq, Nq, Nd, Nd, Nd])
    for q0 in range(Nq):
        for q1 in range(Nq):
            for q2 in range(Nq):
                for i0 in range(Nd):
                    for i1 in range(Nd):
                        for i2 in range(Nd):
                            dphi_tensor[q0, q1, q2, i0, i1, i2] = phi0[q0, i0]*dphi0[q1, i1]*phi0[q2, i2]

    dphi_tensor = dphi_tensor.reshape([Nq*Nq*Nq, Nd*Nd*Nd])
    assert numpy.allclose(dphi_y[:, perm], dphi_tensor)

    # Compute the derivative of basis function in the z direction
    dphi_tensor = numpy.zeros([Nq, Nq, Nq, Nd, Nd, Nd])
    for q0 in range(Nq):
        for q1 in range(Nq):
            for q2 in range(Nq):
                for i0 in range(Nd):
                    for i1 in range(Nd):
                        for i2 in range(Nd):
                            dphi_tensor[q0, q1, q2, i0, i1, i2] = phi0[q0, i0]*phi0[q1, i1]*dphi0[q2, i2]

    dphi_tensor = dphi_tensor.reshape([Nq*Nq*Nq, Nd*Nd*Nd])
    assert numpy.allclose(dphi_z[:, perm], dphi_tensor)



# Copyright (c) 2021 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import os
import pytest
import sys


def test_generated_docs():
    # If this test fails, you should run `python generate_docs.py` in the
    # python/ folder to re-generate the docs for the pybind interface

    path = os.path.dirname(os.path.realpath(__file__))
    if not os.path.isfile(os.path.join(path, "../python/docs.h")):
        pytest.skip("This test can only be run from the source directory.")

    with open(os.path.join(path, "../python/docs.h")) as f:
        docs = f.read()

    sys.path.append(os.path.join(path, "../python"))
    from generate_docs import generate_docs
    assert generate_docs() == docs


def test_demo_index():
    path = os.path.dirname(os.path.realpath(__file__))
    if not os.path.isfile(os.path.join(path, "../demo/python/index.rst")):
        pytest.skip("This test can only be run from the source directory.")

    with open(os.path.join(path, "../demo/python/index.rst")) as f:
        index = f.read()

    for file in os.listdir(os.path.join(path, "../demo/python")):
        if file.endswith(".py") and file.startswith("demo_"):
            assert file in index



import os
import subprocess
import sys
import sysconfig

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

if sys.version_info < (3, 7):
    print("Python 3.7 or higher required, please upgrade.")
    sys.exit(1)

CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Operating System :: POSIX
Operating System :: POSIX :: Linux
Operating System :: MacOS :: MacOS X
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Topic :: Scientific/Engineering :: Mathematics
Topic :: Software Development :: Libraries :: Python Modules
"""


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: "
                               + ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPython3_EXECUTABLE=' + sys.executable,
                      '-DPython3_LIBRARIES=' + sysconfig.get_config_var("LIBDEST"),
                      '-DPython3_INCLUDE_DIRS=' + sysconfig.get_config_var("INCLUDEPY")]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
        build_args += ['--', '-j3']

        env = os.environ.copy()
        import pybind11
        env['pybind11_DIR'] = pybind11.get_cmake_dir()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp, env=env)


setup(name='fenics-basix',
      version='0.4.2.post1',
      description='Basix Python interface',
      url="https://github.com/FEniCS/basix",
      author='FEniCS Project',
      author_email="fenics-dev@googlegroups.com",
      maintainer_email="fenics-dev@googlegroups.com",
      license="MIT License",
      classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
      platforms=["Linux", "Mac OS-X", "Unix"],
      packages=["basix"],
      setup_requires=["pybind11"],
      ext_modules=[CMakeExtension('basix._basixcpp')],
      cmdclass=dict(build_ext=CMakeBuild),
      zip_safe=False)



"""Utility to update the version number in all locations.

Example usage
-------------
To update the version numbers in all files to "1.0.0", run either of the following:
```bash
python3 update_version.py -v 1.0.0
python3 update_version.py --version 1.0.0
```

To update the C++ version numbers to "1.0.0.3" and the Python version numbers to "1.0.0dev3", run
either of the following:
```bash
python3 update_version.py -v 1.0.0.dev3
python3 update_version.py --version 1.0.0.dev3
```
"""

import argparse
import os
import re


def replace_version(content, version):
    content = re.sub(r"((?:VERSION)|(?:version))([\s=]+)([\"']).+?\3",
                     lambda matches: f"{matches[1]}{matches[2]}{matches[3]}{version}{matches[3]}", content)
    content = re.sub(r"(\s+)(\"?)fenics-((?:basix)|(?:ffcx)|(?:dolfinx))\>\=.+(\2|\n)",
                     lambda matches: f"{matches[1]}{matches[2]}fenics-{matches[3]}>={version},<{next_version}{matches[4]}",
                     content)
    return content


parser = argparse.ArgumentParser(description="Update version numbering")
parser.add_argument('-v', '--version', metavar='version', help="Version number to update to", required=True)

args = parser.parse_args()

version = args.version
pyversion = version
if ".dev" in version:
    version = version.replace("dev", "")

print("About to update version numbers to:")
print(f"    C++ version: {version}")
print(f"    Python version: {pyversion}")

answer = None
while answer not in ["Y", "N"]:
    if answer is None:
        answer = input("Do you want to proceed? [Y/N] ").upper()
    else:
        answer = input("Please enter Y or N: ").upper()

if answer == "N":
    print("Aborting.")
    exit()


path = os.path.dirname(os.path.realpath(__file__))

for file in ["CMakeLists.txt", "cpp/CMakeLists.txt", "python/CMakeLists.txt"]:
    print(f"Replacing version numbers in {file}.")
    with open(os.path.join(path, file)) as f:
        content = f.read()
    with open(os.path.join(path, file), "w") as f:
        f.write(replace_version(content, version))

for file in ["setup.py", "python/setup.py"]:
    print(f"Replacing version numbers in {file}.")
    with open(os.path.join(path, file)) as f:
        content = f.read()
    with open(os.path.join(path, file), "w") as f:
        f.write(replace_version(content, pyversion))



"""Basix is a finite element definition and tabulation runtime library.

The core of the library is written in C++, but the majority of Basix's
functionality can be used via this Python interface.
"""

from ._basixcpp import __version__
from . import cell, finite_element, lattice, quadrature, variants
from ._basixcpp import (CellType, LatticeType, LatticeSimplexMethod, ElementFamily, LagrangeVariant,
                        DPCVariant, QuadratureType, PolynomialType, MapType)
from ._basixcpp import (create_lattice, create_element, compute_interpolation_operator, topology,
                        geometry, make_quadrature, index, tabulate_polynomials, create_custom_element)



# =========================
# Creating a custom element
# =========================
#
# In Basix, it is possible to create custom finite elements. This demo describes
# how to do this and what data you need to provide when creating a custom element.
#
# First, we import Basix and Numpy.

import basix
import numpy as np
from basix import CellType, MapType, PolynomialType, LatticeType

# Lagrange element with bubble
# ============================
#
# As a first example, we create a degree 1 Lagrange element with a quadratic bubble
# on a quadrilateral cell. This element will span the following set of polynomials:
#
# .. math::
#    \left\{1,\; x,\; y,\; xy,\; x(1-x)y(1-y)\right\}.
#
# We will define the degrees of freedom (DOFs) of this element by placing a point
# evaluation at each vertex, plus one at the midpoint of the cell.
#
# Polynomial coefficients
# -----------------------
#
# When creating a custom element, we must input the coefficients that define
# a basis of the set of polynomials that our element spans. In this example,
# we will represent the 5 functions above in terms of the 9 orthogonal polynomials
# of degree :math:`\leqslant2` on a quadrilateral, so we create a 5 by 9 matrix.

wcoeffs = np.zeros((5, 9))

# The degree 3 orthonormal polynomials for a quadrilateral will have their highest
# degree terms in the following order:
#
# .. math::
#   1,\; y,\; y^2,\; x,\; xy,\; xy^2,\; x^2,\; x^2y,\; x^2y^2
#
# The order in which the polynomials appear in the orthonormal polynomial sets for
# each cell are documented at
# https://docs.fenicsproject.org/basix/main/polyset-order.html.
#
# As our polynomial space contains 1, :math:`y`, :math:`x` and :math:`xy`. The first
# four rows of the matrix contain a single 1 for the four orthogonal polynomials with
# these are their highest degree terms.

wcoeffs[0, 0] = 1
wcoeffs[1, 1] = 1
wcoeffs[2, 3] = 1
wcoeffs[3, 4] = 1

# The final row of the matrix defines the polynomials :math:`x(1-x)y(1-y)`. As the polynomials
# are orthonormal, we can represent this as
#
# .. math::
#    x(1-x)y(1-y) = \sum_{i=0}^8\int_0^1\int_0^1p_i(x, y)x(1-x)y(1-y)\,\mathrm{d}x\,\mathrm{d}y\; p_i(x, y),
#
# where $p_0$ to $p_8$ are the orthonormal polynomials. Therefore the coefficients we want
# to put in the final row of our matrix are:
#
# .. math::
#    \int_0^1\int_0^1p_i(x, y)x(1-x)y(1-y)\,\mathrm{d}x\,\mathrm{d}y.
#
# We compute these integrals using a degree 4 quadrature rule (this is the largest degree
# that the integrand will be, so these integrals will be exact).

pts, wts = basix.make_quadrature(CellType.quadrilateral, 4)
poly = basix.tabulate_polynomials(PolynomialType.legendre, CellType.quadrilateral, 2, pts)
x = pts[:, 0]
y = pts[:, 1]
f = x * (1 - x) * y * (1 - y)
for i in range(9):
    wcoeffs[4, i] = sum(f * poly[:, i] * wts)

# Interpolation
# -------------
#
# Next, we compute the points and matrices that define how functions can be interpolated
# into this space. These are representations of the functionals that are used in the
# Ciarlet definition of the finite element -- in this example, these are evaluations
# at the points described above.
#
# First, we define the points. We create an array of points for each entity of each
# dimension. For each vertex of the cell, we include the coordinates of that vertex.
# For the interior of the cell, we include a point at :math:`(0.5,0.5)`.
#
# The shape of each of the point lists is (number of points, dimension).

x = [[], [], [], []]
x[0].append(np.array([[0.0, 0.0]]))
x[0].append(np.array([[1.0, 0.0]]))
x[0].append(np.array([[0.0, 1.0]]))
x[0].append(np.array([[1.0, 1.0]]))
x[2].append(np.array([[0.5, 0.5]]))

# There are no DOFs associates with the edges for this element, so we add an empty
# array of points for each edge.

for _ in range(4):
    x[1].append(np.zeros((0, 2)))

# We then define the interpolation matrices that define how the evaluations at the points
# are combined to evaluate the functionals. As all the DOFs are point evaluations in this
# example, the matrices are all identity matrices for the entities that have a point.
#
# The shape of each matrix is (number of DOFs, value size, number of points).

M = [[], [], [], []]
for _ in range(4):
    M[0].append(np.array([[[1.]]]))
M[2].append(np.array([[[1.]]]))

# There are no DOFs associates with the edges for this element, so we add an empty
# matrix for each edge.

for _ in range(4):
    M[1].append(np.zeros((0, 1, 0)))

# Creating the element
# --------------------
#
# We now create the custom element. The inputs into `basix.create_custom_element` are:
#
# - The cell type. In this example, this is a quadrilateral.
# - The value shape of the element. In this example, this is `[]` as the element is scalar.
# - The coefficients that define the polynomial set. In this example, this is `wcoeffs`.
# - The points used to define interpolation into the element. In this example, this is `x`.
# - The matrix used to define interpolation into the element. In this example, this is `M`.
# - The map type. In this example, this is the identity map.
# - A bool indicating whether the element is discontinuous. In this example, this is `False`.
# - The highest degree :math:`n` such that all degree :math:`n` polynomials are contained in
#   this set. In this example, this is 1.
# - The highest degree of a polynomial in the element. In this example, this is 2. It is
#   important that this value is correct, as it will be used to determine the number of
#   polynomials to use when creating and tabulating the element.

element = basix.create_custom_element(
    CellType.quadrilateral, [], wcoeffs, x, M, MapType.identity, False, 1, 2)

# We can now use this element in the same way we can use a built-in element. For example, we
# can tabulate the element at a set of points. If the points we use are the same as the points
# we used to define the DOFs, we see that each basis function is equal to 1 at one of these points,
# and equal to zero at all the other points.

points = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0], [0.5, 0.5]])
print(element.tabulate(0, points))

# Degree 1 Ravairt--Thomas element
# ================================
#
# As a second example, we create a degree 1 Raviart--Thomas element on a triangle. Details of
# the definition of this element can be found at
# https://defelement.com/elements/raviart-thomas.html. This element
# spans:
#
# .. math::
#    \left\{(1, 0),\; (0, 1),\; (x, y)\right\}.
#
# The DOFs of this element are integrals along each edge of the cell of the dot product of the
# function with the normal to the edge.
#
# In contrast to the scalar-valued element above, this element has a value size of 2.
#
# Polynomial coefficients
# -----------------------
#
# In this example, we will represent the 3 functions above in terms of the 3 orthogonal
# polynomials of degree :math:`\leqslant1` on a triangle in each of the two coordinate directions.
# We therefore create a 3 by 6 matrix.

wcoeffs = np.zeros((3, 6))

# The highest degree terms in each polynomial will be:
#
# .. math::
#   (1, 0),\; (x, 0),\; (y, 0),\; (0, 1),\; (0, x),\; (0, y)
#
# We include :math:`(1,0)` and :math:`(0,1)` as the first two rows of the matrix, and use
# integrals to represent :math:`(x,y)` as in the previous example.

wcoeffs[0, 0] = 1
wcoeffs[1, 3] = 1

pts, wts = basix.make_quadrature(CellType.triangle, 2)
poly = basix.tabulate_polynomials(PolynomialType.legendre, CellType.triangle, 1, pts)
x = pts[:, 0]
y = pts[:, 1]
for i in range(3):
    wcoeffs[2, i] = sum(x * poly[:, i] * wts)
    wcoeffs[2, 3 + i] = sum(y * poly[:, i] * wts)

# Interpolation
# -------------
#
# For this element, there will be multiple points used per DOF, as the functionals that define
# the element are integrals. We begin by defining a degree 1 quadrature rule on an interval.
# This quadrature rule will be used to integrate on the edges of the triangle.

pts, wts = basix.make_quadrature(CellType.interval, 1)

# The points associated with each edge are calculated by mapping the quadrature points to each edge.

x = [[], [], [], []]
for _ in range(3):
    x[0].append(np.zeros((0, 2)))
x[1].append(np.array([[1 - p[0], p[0]] for p in pts]))
x[1].append(np.array([[0, p[0]] for p in pts]))
x[1].append(np.array([[p[0], 0] for p in pts]))
x[2].append(np.zeros((0, 2)))

# The interpolation matrices for the edges in this example will be have shape (1, 2, len(pts)), as
# there is one DOF per edge, the value size is 2, and we have len(pts) quadrature points on each
# edge. The entries of these matrices are the quadrature weights multiplied by the normal directions.

M = [[], [], [], []]
for _ in range(3):
    M[0].append(np.zeros((0, 2, 0)))
for normal in [[-1, -1], [-1, 0], [0, 1]]:
    M[1].append(np.array([[normal[0] * wts, normal[1] * wts]]))
M[2].append(np.zeros((0, 2, 0)))

# Creating the element
# --------------------

element = basix.create_custom_element(
    CellType.triangle, [2], wcoeffs, x, M, MapType.contravariantPiola, False, 0, 1)

# To confirm that we have defined this element correctly, we compare it to the built-in
# Raviart--Thomas element.

rt = basix.create_element(basix.ElementFamily.RT, CellType.triangle, 1)

points = basix.create_lattice(CellType.triangle, 1, LatticeType.equispaced, True)
assert np.allclose(rt.tabulate(0, points), element.tabulate(0, points))



# Copyright (c) 2020 Chris Richardson & Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import pytest
import basix
import numpy as np


@pytest.mark.parametrize("lattice_type, simplex_method", [
    (basix.LatticeType.equispaced, basix.LatticeSimplexMethod.none),
    # (basix.LatticeType.gll, basix.LatticeSimplexMethod.warp),
])
@pytest.mark.parametrize("n", [1, 2, 4, 8])
def test_pyramid(n, lattice_type, simplex_method):
    # Check that all the surface points of the pyramid match up with the
    # same points on quad and triangle
    tri_pts = basix.create_lattice(basix.CellType.triangle, n, lattice_type, True, simplex_method)
    quad_pts = basix.create_lattice(basix.CellType.quadrilateral, n, lattice_type, True, simplex_method)
    pyr_pts = basix.create_lattice(basix.CellType.pyramid, n, lattice_type, True, simplex_method)

    # Remove any near-zero values to make sorting robust
    pyr_pts[np.where(abs(pyr_pts) < 1e-12)] = 0.0
    tri_pts[np.where(abs(tri_pts) < 1e-12)] = 0.0
    quad_pts[np.where(abs(quad_pts) < 1e-12)] = 0.0

    idx = np.where(np.isclose(pyr_pts[:, 0], 0.0))
    pyr_x0 = pyr_pts[idx][:, 1:]
    assert np.allclose(np.sort(tri_pts), np.sort(pyr_x0))

    idx = np.where(np.isclose(pyr_pts[:, 0] + pyr_pts[:, 2], 1.0))
    pyr_xz = pyr_pts[idx][:, 1:]
    assert np.allclose(np.sort(tri_pts), np.sort(pyr_xz))

    idx = np.where(np.isclose(pyr_pts[:, 1], 0.0))
    pyr_y0 = pyr_pts[idx][:, 0::2]
    assert np.allclose(np.sort(tri_pts), np.sort(pyr_y0))

    idx = np.where(np.isclose(pyr_pts[:, 1] + pyr_pts[:, 2], 1.0))
    pyr_yz = pyr_pts[idx][:, 0::2]
    assert np.allclose(np.sort(tri_pts), np.sort(pyr_yz))

    idx = np.where(np.isclose(pyr_pts[:, 2], 0.0))
    pyr_z0 = pyr_pts[idx][:, :2]
    assert np.allclose(np.sort(quad_pts), np.sort(pyr_z0))


@pytest.mark.parametrize("lattice_type, simplex_method", [
    (basix.LatticeType.equispaced, basix.LatticeSimplexMethod.none),
    (basix.LatticeType.gll, basix.LatticeSimplexMethod.warp),
    (basix.LatticeType.gll, basix.LatticeSimplexMethod.isaac),
])
@pytest.mark.parametrize("n", [1, 2, 4, 8])
def test_tetrahedron(n, lattice_type, simplex_method):
    # Check that all the surface points of the tet match up with the same points on
    # triangle
    tri_pts = basix.create_lattice(basix.CellType.triangle, n, lattice_type, True, simplex_method)
    tet_pts = basix.create_lattice(basix.CellType.tetrahedron, n, lattice_type, True, simplex_method)

    tet_pts[np.where(abs(tet_pts) < 1e-12)] = 0.0
    tri_pts[np.where(abs(tri_pts) < 1e-12)] = 0.0

    idx = np.where(np.isclose(tet_pts[:, 0], 0.0))
    tet_x0 = tet_pts[idx][:, 1:]
    assert np.allclose(np.sort(tri_pts), np.sort(tet_x0))

    idx = np.where(np.isclose(tet_pts[:, 1], 0.0))
    tet_y0 = tet_pts[idx][:, 0::2]
    assert np.allclose(np.sort(tri_pts), np.sort(tet_y0))

    idx = np.where(np.isclose(tet_pts[:, 2], 0.0))
    tet_z0 = tet_pts[idx][:, :2]
    assert np.allclose(np.sort(tri_pts), np.sort(tet_z0))

    # Project x+y+z=1 onto x=0
    idx = np.where(np.isclose(tet_pts[:, 0] + tet_pts[:, 1] + tet_pts[:, 2], 1.0))
    tet_xyz = tet_pts[idx][:, 1:]
    assert np.allclose(np.sort(tri_pts), np.sort(tet_xyz))


@pytest.mark.parametrize("lattice_type, simplex_method", [
    (basix.LatticeType.equispaced, basix.LatticeSimplexMethod.none),
    (basix.LatticeType.gll, basix.LatticeSimplexMethod.warp),
    (basix.LatticeType.gll, basix.LatticeSimplexMethod.isaac),
])
@pytest.mark.parametrize("n", [1, 2, 4, 8])
def test_triangle(n, lattice_type, simplex_method):
    # Check that all the surface points of the triangle match up with the same points on
    # an interval
    tri_pts = basix.create_lattice(basix.CellType.triangle, n, lattice_type, True, simplex_method)
    interval_pts = basix.create_lattice(basix.CellType.interval, n, lattice_type, True, simplex_method)

    tri_pts[np.where(abs(tri_pts) < 1e-12)] = 0.0
    interval_pts[np.where(abs(interval_pts) < 1e-12)] = 0.0

    idx = np.where(np.isclose(tri_pts[:, 0], 0.0))
    tri_x0 = tri_pts[idx][:, 1:]
    assert np.allclose(np.sort(interval_pts), np.sort(tri_x0))

    idx = np.where(np.isclose(tri_pts[:, 1], 0.0))
    tri_y0 = tri_pts[idx][:, :1]
    assert np.allclose(np.sort(interval_pts), np.sort(tri_y0))

    # Project x+y=1 onto x=0
    idx = np.where(np.isclose(tri_pts[:, 0] + tri_pts[:, 1], 1.0))
    tri_xyz = tri_pts[idx][:, 1:]
    assert np.allclose(np.sort(interval_pts), np.sort(tri_xyz))



# Copyright (c) 2020 Chris Richardson
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import pytest
import numpy as np


@pytest.mark.parametrize("degree", range(1, 4))
@pytest.mark.parametrize("cell", [basix.CellType.triangle, basix.CellType.tetrahedron])
def test_discontinuous_regge(degree, cell):
    e = basix.create_element(basix.ElementFamily.Regge, cell, degree)
    d_e = basix.create_element(basix.ElementFamily.Regge, cell, degree, True)

    pts = basix.create_lattice(cell, 5, basix.LatticeType.equispaced, True)

    assert np.allclose(e.tabulate(1, pts), d_e.tabulate(1, pts))

    for dofs in d_e.num_entity_dofs[:-1]:
        for d in dofs:
            assert d == 0



import basix
import pytest
import numpy as np

elements = [
    (basix.ElementFamily.P, [basix.LagrangeVariant.equispaced]),
    (basix.ElementFamily.P, [basix.LagrangeVariant.gll_warped]),
    (basix.ElementFamily.RT, []),
    (basix.ElementFamily.BDM, []),
    (basix.ElementFamily.N1E, []),
    (basix.ElementFamily.N2E, []),
    (basix.ElementFamily.Regge, []),
    (basix.ElementFamily.HHJ, []),
    (basix.ElementFamily.bubble, []),
    (basix.ElementFamily.serendipity, [basix.LagrangeVariant.legendre, basix.DPCVariant.legendre]),
    (basix.ElementFamily.DPC, [basix.DPCVariant.legendre]),
    (basix.ElementFamily.CR, []),
]


def cross2d(x):
    return [x[1], -x[0]]


def create_continuity_map_interval(map_type, start, end):
    if map_type == basix.MapType.identity:
        return lambda x: x
    if map_type == basix.MapType.covariantPiola:
        return lambda x: np.dot(x, end - start)
    if map_type == basix.MapType.contravariantPiola:
        return lambda x: np.dot(x, cross2d(end - start))
    if map_type == basix.MapType.doubleCovariantPiola:
        return lambda x: np.dot(start - end, np.dot(x, end - start))
    if map_type == basix.MapType.doubleContravariantPiola:
        return lambda x: np.dot(cross2d(end - start), np.dot(x, cross2d(end - start)))

    raise NotImplementedError


def create_continuity_map_triangle(map_type, v0, v1, v2):
    if map_type == basix.MapType.identity:
        return lambda x: x

    raise NotImplementedError


def create_continuity_map_quadrilateral(map_type, v0, v1, v2):
    if map_type == basix.MapType.identity:
        return lambda x: x

    raise NotImplementedError


@pytest.mark.parametrize("degree", range(1, 5))
@pytest.mark.parametrize("element, variant", elements)
def test_continuity_interval_facet(degree, element, variant):
    """Test that basis functions between neighbouring cells of different types will be continuous."""
    elements = {}
    for cell in [basix.CellType.triangle, basix.CellType.quadrilateral]:
        try:
            elements[cell] = basix.create_element(element, cell, degree, *variant)
        except RuntimeError:
            pass

    if len(elements) <= 1:
        pytest.skip()

    facets = [
        [np.array([0, 0]), np.array([1, 0]), {basix.CellType.triangle: 2, basix.CellType.quadrilateral: 0}],
        [np.array([0, 0]), np.array([0, 1]), {basix.CellType.triangle: 1, basix.CellType.quadrilateral: 1}],
    ]

    for start, end, cellmap in facets:
        points = np.array([start + i/10 * (end - start) for i in range(11)])

        data = None

        for c, e in elements.items():
            tab = e.tabulate(0, points)[0]
            continuity_map = create_continuity_map_interval(e.map_type, start, end)
            entity_tab = [continuity_map(tab[:, i, :]) for i in e.entity_dofs[1][cellmap[c]]]
            if data is None:
                data = entity_tab
            else:
                assert np.allclose(data, entity_tab)


@pytest.mark.parametrize("degree", range(1, 5))
@pytest.mark.parametrize("element, variant", elements)
def test_continuity_triangle_facet(degree, element, variant):
    """Test that basis functions between neighbouring cells of different types will be continuous."""
    elements = {}
    for cell in [basix.CellType.tetrahedron, basix.CellType.prism]:  # , basix.CellType.pyramid]:
        try:
            elements[cell] = basix.create_element(element, cell, degree, *variant)
        except RuntimeError:
            pass

    if len(elements) <= 1:
        pytest.skip()

    facets = [
        [
            np.array([0, 0, 0]), np.array([1, 0, 0]), np.array([0, 1, 0]),
            {basix.CellType.tetrahedron: 3, basix.CellType.prism: 0}
        ], [
            np.array([0, 0, 0]), np.array([1, 0, 0]), np.array([0, 0, 1]),
            {basix.CellType.tetrahedron: 2, basix.CellType.pyramid: 1}
        ], [
            np.array([0, 0, 0]), np.array([0, 1, 0]), np.array([0, 0, 1]),
            {basix.CellType.tetrahedron: 1, basix.CellType.pyramid: 2}
        ],
    ]

    for v0, v1, v2, cellmap in facets:
        points = np.array([v0 + i/10 * (v1 - v0) + j/10 * (v2 - v0) for i in range(11) for j in range(11 - i)])

        data = None

        for c, e in elements.items():
            if c in cellmap:
                tab = e.tabulate(0, points)
                continuity_map = create_continuity_map_triangle(e.map_type, v0, v1, v2)
                entity_tab = [continuity_map(tab[:, :, i, :]) for i in e.entity_dofs[2][cellmap[c]]]
                if data is None:
                    data = entity_tab
                else:
                    assert np.allclose(data, entity_tab)


@pytest.mark.parametrize("degree", range(1, 5))
@pytest.mark.parametrize("element, variant", elements)
def test_continuity_quadrilateral_facet(degree, element, variant):
    """Test that basis functions between neighbouring cells of different types will be continuous."""
    elements = {}
    for cell in [basix.CellType.hexahedron, basix.CellType.prism]:  # , basix.CellType.pyramid]:
        try:
            elements[cell] = basix.create_element(element, cell, degree, *variant)
        except RuntimeError:
            pass

    if len(elements) <= 1:
        pytest.skip()

    facets = [
        [
            np.array([0, 0, 0]), np.array([1, 0, 0]), np.array([0, 1, 0]), np.array([1, 1, 0]),
            {basix.CellType.hexahedron: 0, basix.CellType.pyramid: 0}
        ], [
            np.array([0, 0, 0]), np.array([1, 0, 0]), np.array([0, 0, 1]), np.array([1, 0, 1]),
            {basix.CellType.hexahedron: 1, basix.CellType.prism: 1}
        ], [
            np.array([0, 0, 0]), np.array([0, 1, 0]), np.array([0, 0, 1]), np.array([0, 1, 1]),
            {basix.CellType.hexahedron: 2, basix.CellType.prism: 2}
        ],
    ]

    for v0, v1, v2, v3, cellmap in facets:
        assert np.allclose(v0 + v3, v1 + v2)
        points = np.array([v0 + i/10 * (v1 - v0) + j/10 * (v2 - v0) for i in range(11) for j in range(11)])

        data = None

        for c, e in elements.items():
            if c in cellmap:
                tab = e.tabulate(0, points)
                continuity_map = create_continuity_map_quadrilateral(e.map_type, v0, v1, v2)
                entity_tab = [continuity_map(tab[:, :, i, :]) for i in e.entity_dofs[2][cellmap[c]]]
                if data is None:
                    data = entity_tab
                else:
                    assert np.allclose(data, entity_tab)



"""Functions to manipulate lattice types."""

from ._basixcpp import LatticeType as _LT
from ._basixcpp import LatticeSimplexMethod as _LSM


def string_to_type(lattice: str) -> _LT:
    """Convert a string to a Basix LatticeType enum.

    Args:
        lattice: The lattice type as a string.

    Returns:
        The lattice type
    """
    if not hasattr(_LT, lattice):
        raise ValueError(f"Unknown lattice: {lattice}")
    return getattr(_LT, lattice)


def type_to_string(latticetype: _LT) -> str:
    """Convert a Basix LatticeType enum to a string.

    Args:
        latticetype: The lattice type

    Returns:
        The lattice type as a string.
    """
    return latticetype.name


def string_to_simplex_method(method: str) -> _LSM:
    """Convert a string to a Basix LatticeSimplexMethod enum.

    Args:
        method: The simplex method as a string.

    Returns:
        The simplex method
    """
    if not hasattr(_LSM, method):
        raise ValueError(f"Unknown simplex method: {method}")
    return getattr(_LSM, method)


def simplex_method_to_string(simplex_method: _LSM) -> str:
    """Convert a Basix LatticeSimplexMethod enum to a string.

    Args:
        simplex_method: The simplex method

    Returns:
        The simplex method as a string.
    """
    return simplex_method.name



# Copyright (c) 2021 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import numpy as np
import basix
import pytest
from .utils import parametrize_over_elements


def run_test(lower_element, higher_element, power, value_size):
    l_points = lower_element.points
    l_eval = np.concatenate([
        l_points[:, 0] ** power if i == 0 else 0 * l_points[:, 0]
        for i in range(value_size)
    ])
    l_coeffs = lower_element.interpolation_matrix @ l_eval

    i_m = basix.compute_interpolation_operator(lower_element, higher_element)
    h_coeffs = i_m @ l_coeffs

    h_points = higher_element.points
    h_eval = np.concatenate([
        h_points[:, 0] ** power if i == 0 else 0 * h_points[:, 0]
        for i in range(value_size)
    ])
    h_coeffs2 = higher_element.interpolation_matrix @ h_eval

    assert np.allclose(h_coeffs, h_coeffs2)


@pytest.mark.parametrize("cell_type", [basix.CellType.interval, basix.CellType.triangle, basix.CellType.tetrahedron,
                                       basix.CellType.quadrilateral, basix.CellType.hexahedron, basix.CellType.prism])
@pytest.mark.parametrize("orders", [(1, 2), (2, 4), (4, 5)])
def test_different_order_interpolation_lagrange(cell_type, orders):
    lower_element = basix.create_element(basix.ElementFamily.P, cell_type, orders[0], basix.LagrangeVariant.gll_warped)
    higher_element = basix.create_element(basix.ElementFamily.P, cell_type, orders[1], basix.LagrangeVariant.gll_warped)

    run_test(lower_element, higher_element, orders[0], lower_element.value_size)


@pytest.mark.parametrize("variant1", [basix.LagrangeVariant.equispaced, basix.LagrangeVariant.gll_warped,
                                      basix.LagrangeVariant.gll_isaac])
@pytest.mark.parametrize("variant2", [basix.LagrangeVariant.equispaced, basix.LagrangeVariant.gll_warped,
                                      basix.LagrangeVariant.gll_isaac])
@pytest.mark.parametrize("cell_type", [basix.CellType.interval, basix.CellType.triangle, basix.CellType.tetrahedron,
                                       basix.CellType.quadrilateral, basix.CellType.hexahedron, basix.CellType.prism])
@pytest.mark.parametrize("order", [1, 4])
def test_different_variant_interpolation(cell_type, order, variant1, variant2):
    lower_element = basix.create_element(basix.ElementFamily.P, cell_type, order, variant1)
    higher_element = basix.create_element(basix.ElementFamily.P, cell_type, order, variant2)

    run_test(lower_element, higher_element, order, lower_element.value_size)


@pytest.mark.parametrize("family, args", [
    [basix.ElementFamily.RT, (basix.LagrangeVariant.legendre, )],
    [basix.ElementFamily.N1E, (basix.LagrangeVariant.legendre, )],
    [basix.ElementFamily.BDM, (basix.LagrangeVariant.legendre, basix.DPCVariant.legendre)],
    [basix.ElementFamily.N2E, (basix.LagrangeVariant.legendre, basix.DPCVariant.legendre)],
])
@pytest.mark.parametrize("cell_type", [basix.CellType.triangle, basix.CellType.tetrahedron,
                                       basix.CellType.quadrilateral, basix.CellType.hexahedron])
@pytest.mark.parametrize("orders", [(1, 2), (2, 4), (4, 5)])
def test_different_order_interpolation_vector(family, args, cell_type, orders):
    lower_element = basix.create_element(family, cell_type, orders[0], *args)
    higher_element = basix.create_element(family, cell_type, orders[1], *args)

    run_test(lower_element, higher_element, orders[0] - 1, lower_element.value_size)


@pytest.mark.parametrize("family, args", [[basix.ElementFamily.Regge, tuple()]])
@pytest.mark.parametrize("cell_type", [basix.CellType.triangle, basix.CellType.tetrahedron])
@pytest.mark.parametrize("orders", [(1, 2), (2, 4), (4, 5)])
def test_different_order_interpolation_matrix(family, args, cell_type, orders):
    lower_element = basix.create_element(family, cell_type, orders[0], *args)
    higher_element = basix.create_element(family, cell_type, orders[1], *args)

    run_test(lower_element, higher_element, orders[0] - 1, lower_element.value_size)


@pytest.mark.parametrize("family1, args1", [
    [basix.ElementFamily.RT, (basix.LagrangeVariant.legendre, )],
    [basix.ElementFamily.N1E, (basix.LagrangeVariant.legendre, )],
    [basix.ElementFamily.BDM, (basix.LagrangeVariant.legendre, basix.DPCVariant.legendre)],
    [basix.ElementFamily.N2E, (basix.LagrangeVariant.legendre, basix.DPCVariant.legendre)],
])
@pytest.mark.parametrize("family2, args2", [
    [basix.ElementFamily.RT, (basix.LagrangeVariant.legendre, )],
    [basix.ElementFamily.N1E, (basix.LagrangeVariant.legendre, )],
    [basix.ElementFamily.BDM, (basix.LagrangeVariant.legendre, basix.DPCVariant.legendre)],
    [basix.ElementFamily.N2E, (basix.LagrangeVariant.legendre, basix.DPCVariant.legendre)],
])
@pytest.mark.parametrize("cell_type", [basix.CellType.triangle, basix.CellType.tetrahedron,
                                       basix.CellType.quadrilateral, basix.CellType.hexahedron])
@pytest.mark.parametrize("order", [1, 4])
def test_different_element_interpolation(family1, args1, family2, args2, cell_type, order):
    lower_element = basix.create_element(family1, cell_type, order, *args1)
    higher_element = basix.create_element(family2, cell_type, order, *args2)

    run_test(lower_element, higher_element, order - 1, lower_element.value_size)


@pytest.mark.parametrize("cell_type", [basix.CellType.triangle, basix.CellType.tetrahedron,
                                       basix.CellType.quadrilateral, basix.CellType.hexahedron])
@pytest.mark.parametrize("order", [1, 4])
def test_blocked_interpolation(cell_type, order):
    """Test interpolation of Nedelec's componenets into a Lagrange space."""
    nedelec = basix.create_element(
        basix.ElementFamily.N2E, cell_type, order, basix.LagrangeVariant.legendre, basix.DPCVariant.legendre)
    lagrange = basix.create_element(basix.ElementFamily.P, cell_type, order, basix.LagrangeVariant.gll_isaac)

    n_points = nedelec.points
    if nedelec.value_size == 2:
        n_eval = np.concatenate([n_points[:, 0] ** order, n_points[:, 1] ** order])
    else:
        n_eval = np.concatenate([n_points[:, 0] ** order, 0 * n_points[:, 0], n_points[:, 1] ** order])
    n_coeffs = nedelec.interpolation_matrix @ n_eval

    l_points = lagrange.points
    if nedelec.value_size == 2:
        values = [l_points[:, 0] ** order, l_points[:, 1] ** order]
    else:
        values = [l_points[:, 0] ** order, 0 * l_points[:, 0], l_points[:, 1] ** order]
    l_coeffs = np.empty(lagrange.dim * nedelec.value_size)
    for i, v in enumerate(values):
        l_coeffs[i::nedelec.value_size] = v

    # Test interpolation from Nedelec to blocked Lagrange
    i_m = basix.compute_interpolation_operator(nedelec, lagrange)
    assert np.allclose(l_coeffs, i_m @ n_coeffs)

    # Test interpolation from blocked Lagrange to Nedelec
    i_m = basix.compute_interpolation_operator(lagrange, nedelec)
    assert np.allclose(n_coeffs, i_m @ l_coeffs)


@parametrize_over_elements(5)
def test_degree_bounds(cell_type, degree, element_type, element_args):
    np.random.seed(13)

    element = basix.create_element(element_type, cell_type, degree, *element_args)

    points = basix.create_lattice(cell_type, 10, basix.LatticeType.equispaced, True)
    tab = element.tabulate(0, points)[0]

    # Test that this element's basis functions are contained in Lagrange space with
    # degree element.highest_degree
    coeffs = np.random.rand(element.dim)
    values = np.array([tab[:, :, i] @ coeffs
                       for i in range(element.value_size)])

    if element.highest_degree >= 0:
        # The element being tested should be a subset of this Lagrange space
        lagrange = basix.create_element(
            basix.ElementFamily.P, cell_type, element.highest_degree, basix.LagrangeVariant.equispaced, True)
        lagrange_coeffs = basix.compute_interpolation_operator(element, lagrange) @ coeffs
        lagrange_tab = lagrange.tabulate(0, points)[0]
        lagrange_values = np.array([lagrange_tab[:, :, 0] @ lagrange_coeffs[i::element.value_size]
                                    for i in range(element.value_size)])

        assert np.allclose(values, lagrange_values)

    if element.highest_degree >= 1:
        # The element being tested should be NOT a subset of this Lagrange space
        lagrange = basix.create_element(
            basix.ElementFamily.P, cell_type, element.highest_degree - 1,
            basix.LagrangeVariant.equispaced, True)
        lagrange_coeffs = basix.compute_interpolation_operator(element, lagrange) @ coeffs
        lagrange_tab = lagrange.tabulate(0, points)[0]
        lagrange_values = np.array([lagrange_tab[:, :, 0] @ lagrange_coeffs[i::element.value_size]
                                    for i in range(element.value_size)])

        assert not np.allclose(values, lagrange_values)

    # Test that the basis functions of Lagrange space with degree element.highest_complete_degree
    # are contained in this space

    if element.highest_complete_degree >= 0:
        # This Lagrange space should be a subset to the element being tested
        lagrange = basix.create_element(
            basix.ElementFamily.P, cell_type, element.highest_complete_degree,
            basix.LagrangeVariant.equispaced, True)
        lagrange_coeffs = np.random.rand(lagrange.dim * element.value_size)
        lagrange_tab = lagrange.tabulate(0, points)[0]
        lagrange_values = np.array([lagrange_tab[:, :, 0] @ lagrange_coeffs[i::element.value_size]
                                    for i in range(element.value_size)])

        coeffs = basix.compute_interpolation_operator(lagrange, element) @ lagrange_coeffs
        values = np.array([tab[:, :, i] @ coeffs
                           for i in range(element.value_size)])

        assert np.allclose(values, lagrange_values)

    if element.highest_complete_degree >= -1:
        # This Lagrange space should NOT be a subset to the element being tested
        lagrange = basix.create_element(
            basix.ElementFamily.P, cell_type, element.highest_complete_degree + 1,
            basix.LagrangeVariant.equispaced, True)
        lagrange_coeffs = np.random.rand(lagrange.dim * element.value_size)
        lagrange_tab = lagrange.tabulate(0, points)[0]
        lagrange_values = np.array([lagrange_tab[:, :, 0] @ lagrange_coeffs[i::element.value_size]
                                    for i in range(element.value_size)])

        coeffs = basix.compute_interpolation_operator(lagrange, element) @ lagrange_coeffs
        values = np.array([tab[:, :, i] @ coeffs
                           for i in range(element.value_size)])

        assert not np.allclose(values, lagrange_values)



"""Functions to get cell geometry information and manipulate cell types."""

from ._basixcpp import cell_volume as volume  # noqa: F401
from ._basixcpp import cell_facet_reference_volumes as facet_reference_volumes  # noqa: F401
from ._basixcpp import cell_facet_normals as facet_normals  # noqa: F401
from ._basixcpp import cell_facet_outward_normals as facet_outward_normals  # noqa: F401
from ._basixcpp import cell_facet_orientations as facet_orientations  # noqa: F401
from ._basixcpp import cell_facet_jacobians as facet_jacobians  # noqa: F401
from ._basixcpp import sub_entity_connectivity  # noqa: F401
from ._basixcpp import CellType as _CT


def string_to_type(cell: str) -> _CT:
    """Convert a string to a Basix CellType.

    Args:
        cell: The name of the cell as a string.

    Returns:
        The cell type
    """
    if not hasattr(_CT, cell):
        raise ValueError(f"Unknown cell: {cell}")
    return getattr(_CT, cell)


def type_to_string(celltype: _CT) -> str:
    """Convert a Basix CellType to a string.

    Args:
        celltype: The cell type.

    Returns:
        The name of the cell as a string.
    """
    return celltype.name



# Copyright (c) 2020 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import random
import basix
import pytest
import numpy as np

elements = [
    (basix.ElementFamily.P, [basix.LagrangeVariant.gll_warped]),  # identity
    (basix.ElementFamily.N1E, [basix.LagrangeVariant.legendre]),  # covariant Piola
    (basix.ElementFamily.RT, [basix.LagrangeVariant.legendre]),  # contravariant Piola
    (basix.ElementFamily.Regge, []),  # double covariant Piola
    (basix.ElementFamily.HHJ, []),  # double contravariant Piola
]


def run_map_test(e, J, detJ, K, reference_value_size, physical_value_size):
    tdim = len(basix.topology(e.cell_type)) - 1
    N = 5
    if tdim == 1:
        points = np.array([[i / N] for i in range(N + 1)])
    elif tdim == 2:
        points = np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1 - i)])
    elif tdim == 3:
        points = np.array([[i / N, j / N, k / N]
                           for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1 - i - j)])
    values = e.tabulate(0, points)[0]

    _J = np.array([J for p in points])
    _detJ = np.array([detJ for p in points])
    _K = np.array([K for p in points])

    assert values.shape[1] == e.dim
    assert values.shape[2] == reference_value_size

    mapped = e.push_forward(values, _J, _detJ, _K)
    assert mapped.shape[0] == values.shape[0]
    assert mapped.shape[1] == e.dim
    assert mapped.shape[2] == physical_value_size

    unmapped = e.pull_back(mapped, _J, _detJ, _K)
    assert np.allclose(values, unmapped)


@pytest.mark.parametrize("element_type, element_args", elements)
def test_mappings_2d_to_2d(element_type, element_args):
    e = basix.create_element(element_type, basix.CellType.triangle, 1, *element_args)
    J = np.array([[random.random() + 1, random.random()],
                  [random.random(), random.random()]])
    detJ = np.linalg.det(J)
    K = np.linalg.inv(J)
    run_map_test(e, J, detJ, K, e.value_size, e.value_size)


@pytest.mark.parametrize("element_type, element_args", elements)
def test_mappings_2d_to_3d(element_type, element_args):
    e = basix.create_element(element_type, basix.CellType.triangle, 1, *element_args)

    # Map from (0,0)--(1,0)--(0,1) to (1,0,1)--(2,1,0)--(0,1,1)
    J = np.array([[1., -1.], [1., 1.], [-1., 0.]])
    detJ = np.sqrt(6)
    K = np.array([[0.5, 0.5, 0.], [-0.5, 0.5, 0.]])

    if e.value_size == 1:
        physical_vs = 1
    elif e.value_size == 2:
        physical_vs = 3
    elif e.value_size == 4:
        physical_vs = 9
    run_map_test(e, J, detJ, K, e.value_size, physical_vs)


@pytest.mark.parametrize("element_type, element_args", elements)
def test_mappings_3d_to_3d(element_type, element_args):
    if element_type == basix.ElementFamily.HHJ:
        pytest.xfail("HHJ not implemented on tetrahedra.")
    random.seed(42)
    e = basix.create_element(element_type, basix.CellType.tetrahedron, 1, *element_args)

    J = np.array([[random.random() + 2, random.random(), random.random()],
                  [random.random(), random.random() + 1, random.random()],
                  [random.random(), random.random(), random.random()]])
    detJ = np.linalg.det(J)
    K = np.linalg.inv(J)

    run_map_test(e, J, detJ, K, e.value_size, e.value_size)



# Copyright (c) 2021 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import random

import basix
import numpy as np
import pytest


@pytest.mark.parametrize("cell", [basix.CellType.triangle, basix.CellType.tetrahedron,
                                  basix.CellType.quadrilateral, basix.CellType.hexahedron])
@pytest.mark.parametrize("element, degree, element_args", [
    (basix.ElementFamily.P, 1, [basix.LagrangeVariant.gll_warped]),
    (basix.ElementFamily.P, 3, [basix.LagrangeVariant.gll_warped]),
    (basix.ElementFamily.N1E, 3, [])
])
@pytest.mark.parametrize("block_size", [1, 2, 4])
def test_dof_transformations(cell, element, degree, element_args, block_size):
    try:
        import numba  # noqa: F401
    except ImportError:
        pytest.skip("Numba must be installed to run this test.")

    from basix import numba_helpers
    from numba.core import types
    from numba.typed import Dict

    transform_functions = {
        basix.CellType.triangle: numba_helpers.apply_dof_transformation_triangle,
        basix.CellType.quadrilateral: numba_helpers.apply_dof_transformation_quadrilateral,
        basix.CellType.tetrahedron: numba_helpers.apply_dof_transformation_tetrahedron,
        basix.CellType.hexahedron: numba_helpers.apply_dof_transformation_hexahedron,
        basix.CellType.prism: numba_helpers.apply_dof_transformation_prism,
        basix.CellType.pyramid: numba_helpers.apply_dof_transformation_pyramid
    }

    random.seed(1337)

    e = basix.create_element(element, cell, degree, *element_args)
    data = np.array(list(range(e.dim * block_size)), dtype=np.double)

    for i in range(10):
        cell_info = random.randrange(2 ** 30)

        data1 = data.copy()
        data1 = e.apply_dof_transformation(data1, block_size, cell_info)
        # Numba function does not use blocked data
        data2 = data.copy().reshape(e.dim, block_size)
        # Mapping lists to numba dictionaries
        entity_transformations = Dict.empty(key_type=types.string, value_type=types.float64[:, :, :])
        for i, transformation in e.entity_transformations().items():
            entity_transformations[i] = transformation

        entity_dofs = Dict.empty(key_type=types.int64, value_type=types.int32[:])
        for i, e_dofs in enumerate(e.num_entity_dofs):
            entity_dofs[i] = np.asarray(e_dofs, dtype=np.int32)
        transform_functions[cell](entity_transformations, entity_dofs, data2, cell_info)
        # Reshape numba output for comparison
        data2 = data2.reshape(-1)
        assert np.allclose(data1, data2)


@pytest.mark.parametrize("cell", [basix.CellType.triangle, basix.CellType.tetrahedron,
                                  basix.CellType.quadrilateral, basix.CellType.hexahedron])
@pytest.mark.parametrize("element, degree, element_args", [
    (basix.ElementFamily.P, 1, [basix.LagrangeVariant.gll_warped]),
    (basix.ElementFamily.P, 3, [basix.LagrangeVariant.gll_warped]),
    (basix.ElementFamily.N1E, 3, [])
])
@pytest.mark.parametrize("block_size", [1, 2, 4])
def test_dof_transformations_to_transpose(cell, element, degree, block_size, element_args):
    try:
        import numba  # noqa: F401
    except ImportError:
        pytest.skip("Numba must be installed to run this test.")

    from basix import numba_helpers
    from numba.core import types
    from numba.typed import Dict

    transform_functions = {
        basix.CellType.triangle: numba_helpers.apply_dof_transformation_to_transpose_triangle,
        basix.CellType.quadrilateral: numba_helpers.apply_dof_transformation_to_transpose_quadrilateral,
        basix.CellType.tetrahedron: numba_helpers.apply_dof_transformation_to_transpose_tetrahedron,
        basix.CellType.hexahedron: numba_helpers.apply_dof_transformation_to_transpose_hexahedron,
        basix.CellType.prism: numba_helpers.apply_dof_transformation_to_transpose_prism,
        basix.CellType.pyramid: numba_helpers.apply_dof_transformation_to_transpose_pyramid
    }

    random.seed(1337)

    e = basix.create_element(element, cell, degree, *element_args)
    data = np.array(list(range(e.dim * block_size)), dtype=np.double)

    for i in range(10):
        cell_info = random.randrange(2 ** 30)

        data1 = data.copy()
        data1 = e.apply_dof_transformation_to_transpose(data1, block_size, cell_info)
        # Numba function does not use blocked data
        data2 = data.copy().reshape(block_size, e.dim)
        # Mapping lists to numba dictionaries
        entity_transformations = Dict.empty(key_type=types.string, value_type=types.float64[:, :, :])
        for i, transformation in e.entity_transformations().items():
            entity_transformations[i] = transformation

        entity_dofs = Dict.empty(key_type=types.int64, value_type=types.int32[:])
        for i, e_dofs in enumerate(e.num_entity_dofs):
            entity_dofs[i] = np.asarray(e_dofs, dtype=np.int32)
        transform_functions[cell](entity_transformations, entity_dofs, data2, cell_info)
        # Reshape numba output for comparison
        data2 = data2.reshape(-1)
        assert np.allclose(data1, data2)



# ==================================
# Creating and tabulating an element
# ==================================
#
# This demo shows how Basix can be used to create an element
# and tabulate the values of its basis functions at a set of
# points
#
# First, we import Basix and Numpy.

import basix
import numpy as np
from basix import ElementFamily, CellType, LagrangeVariant

# Next, we create a degree 4 Lagrange element on a quadrilateral using the function
# `create_element`. The first input is the element family: for Lagrange elements,
# we use `ElementFamily.P`. The second input is the cell type. The third
# input is the degree of the element. For Lagrange elements, we must provide a
# fourth input: the Lagrange variant. In this example, we use the equispaced
# variant: this will place the degrees of freedom (DOFs) of the element in an
# equally spaced lattice.

lagrange = basix.create_element(
    ElementFamily.P, CellType.quadrilateral, 4, LagrangeVariant.equispaced)

# We now print the number of DOFs that this element has.

print(lagrange.dim)

# We see that the element has 25 DOFs: for this element, there will be 1 DOFs at
# each vertex, 3 DOFs on each edge, and 9 DOFs on the interior of the quadrilateral.
#
# Next, we create a set of points as a numpy array, and tabulate the basis functions
# of the Lagrange space at these points. The first input of `tabulate` is the number
# of derivative to tabulate: we set this to 0 so will to compute the values of the
# functions (and no derivatives). We pass in the points as the second input.

points = np.array([[0.0, 0.0], [0.1, 0.1], [0.2, 0.3], [0.3, 0.6], [0.4, 1.0]])
tab = lagrange.tabulate(0, points)
print(tab)
print(tab.shape)

# The result of tabulating is a 1 by 5 by 25 by 1 Numpy array. The first dimension
# is 1 as we are only tabulating the function values; it would be higher if we
# had asked for derivatives too. The second dimension (5) is the number of points.
# The third dimension (25) is the number of DOFs. The fourth dimension (1) is the
# value size of the element: this will be greater than 1 for vector-values elements.
#
# C++ demo
# ========
# The following C++ code runs the same demo using Basix's C++ interface:
#
# .. literalinclude:: ../cpp/demo_create_and_tabulate/main.cpp
#    :language: c++



# Copyright (c) 2020 Chris Richardson
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import numpy as np
import pytest


@pytest.mark.parametrize("order", [1, 2, 3])
def test_quad(order):
    Lpts, Lwts = basix.make_quadrature(basix.CellType.interval, 2*order + 1)
    Qwts = []
    Qpts = []
    for p, u in zip(Lpts, Lwts):
        for q, v in zip(Lpts, Lwts):
            Qpts.append([p[0], q[0]])
            Qwts.append(u * v)
    basis = basix._basixcpp.tabulate_polynomial_set(basix.CellType.quadrilateral,
                                                    order, 0, Qpts)[0]
    ndofs = basis.shape[1]

    mat = np.zeros((ndofs, ndofs))
    for i in range(ndofs):
        for j in range(ndofs):
            mat[i, j] = sum(basis[:, i] * basis[:, j] * Qwts)

    assert np.allclose(mat, np.eye(ndofs))


@pytest.mark.parametrize("order", [1, 2, 3, 4])
def test_pyramid(order):
    Lpts, Lwts = basix.make_quadrature(basix.CellType.interval, 4*order + 2)
    Qwts = []
    Qpts = []
    for p, u in zip(Lpts, Lwts):
        for q, v in zip(Lpts, Lwts):
            for r, w in zip(Lpts, Lwts):
                sc = (1.0 - r[0])
                Qpts.append([p[0] * sc, q[0] * sc, r[0]])
                Qwts.append(u * v * sc * sc * w)
    basis = basix._basixcpp.tabulate_polynomial_set(basix.CellType.pyramid,
                                                    order, 0, Qpts)[0]
    ndofs = basis.shape[1]

    mat = np.zeros((ndofs, ndofs))
    for i in range(ndofs):
        for j in range(ndofs):
            mat[i, j] = sum(basis[:, i] * basis[:, j] * Qwts)

    assert np.allclose(mat, np.eye(ndofs))


@pytest.mark.parametrize("order", [1, 2, 3])
def test_hex(order):
    Lpts, Lwts = basix.make_quadrature(basix.CellType.interval, 2*order + 1)
    Qwts = []
    Qpts = []
    for p, u in zip(Lpts, Lwts):
        for q, v in zip(Lpts, Lwts):
            for r, w in zip(Lpts, Lwts):
                Qpts.append([p[0], q[0], r[0]])
                Qwts.append(u * v * w)
    basis = basix._basixcpp.tabulate_polynomial_set(basix.CellType.hexahedron,
                                                    order, 0, Qpts)[0]
    ndofs = basis.shape[1]

    mat = np.zeros((ndofs, ndofs))
    for i in range(ndofs):
        for j in range(ndofs):
            mat[i, j] = sum(basis[:, i] * basis[:, j] * Qwts)

    assert np.allclose(mat, np.eye(ndofs))


@pytest.mark.parametrize("order", [1, 2, 3])
def test_prism(order):
    Tpts, Twts = basix.make_quadrature(basix.CellType.triangle, 2*order + 1)
    Lpts, Lwts = basix.make_quadrature(basix.CellType.interval, 2*order + 1)
    Qwts = []
    Qpts = []
    for p, u in zip(Tpts, Twts):
        for q, v in zip(Lpts, Lwts):
            Qpts.append([p[0], p[1], q[0]])
            Qwts.append(u * v)
    basis = basix._basixcpp.tabulate_polynomial_set(basix.CellType.prism,
                                                    order, 0, Qpts)[0]
    ndofs = basis.shape[1]

    mat = np.zeros((ndofs, ndofs))
    for i in range(ndofs):
        for j in range(ndofs):
            mat[i, j] = sum(basis[:, i] * basis[:, j] * Qwts)

    assert np.allclose(mat, np.eye(ndofs))


@pytest.mark.parametrize("cell_type", [
    basix.CellType.interval,
    basix.CellType.triangle,
    basix.CellType.quadrilateral,
    basix.CellType.tetrahedron,
    basix.CellType.hexahedron,
    basix.CellType.prism,
])
@pytest.mark.parametrize("order", [0, 1, 2, 3, 4])
def test_cell(cell_type, order):
    Qpts, Qwts = basix.make_quadrature(cell_type, 2*order + 1)
    basis = basix._basixcpp.tabulate_polynomial_set(cell_type, order, 0, Qpts)[0]

    ndofs = basis.shape[1]
    mat = np.zeros((ndofs, ndofs))
    for i in range(ndofs):
        for j in range(ndofs):
            mat[i, j] = sum(basis[:, i] * basis[:, j] * Qwts)

    assert np.allclose(mat, np.eye(ndofs))



from skbuild import setup

import sys
import sysconfig

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name="fenics-basix",
      python_requires='>=3.7.0',
      version="0.4.2.post1",
      description='Basix Python interface',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/FEniCS/basix",
      author='FEniCS Project',
      author_email="fenics-dev@googlegroups.com",
      maintainer_email="fenics-dev@googlegroups.com",
      license="MIT",
      packages=["basix"],
      install_requires=["numpy"],
      extras_require={
          "docs": ["markdown", "pylit3", "pyyaml", "sphinx", "sphinx_rtd_theme"],
          "lint": ["flake8", "pydocstyle"],
          "optional": ["numba"],
          "test": ["pytest", "sympy", "numba", "scipy", "matplotlib", "fenics-ufl"],
          "ci": ["pytest-xdist", "fenics-basix[docs]", "fenics-basix[lint]", "fenics-basix[optional]",
                 "fenics-basix[test]"]
      },
      cmake_args=[
          '-DPython3_EXECUTABLE=' + sys.executable,
          '-DPython3_LIBRARIES=' + sysconfig.get_config_var("LIBDEST"),
          '-DPython3_INCLUDE_DIRS=' + sysconfig.get_config_var("INCLUDEPY"),
          '-DDOWNLOAD_XTENSOR_LIBS=ON'],
      package_dir={"": "python"},
      cmake_install_dir="python/basix/")



# ====================================
# DOF permutations and transformations
# ====================================
#
# When using high degree finite elements on general meshes, adjustments
# may need to be made to correct for differences in the orientation of
# mesh entities on the mesh and on the reference cell. For example, in
# a degree 4 Lagrange element on a triangle, there are 3 degrees of
# freedom (DOFs) associated with each edge. If two neighbouring cells in
# a mesh disagree on the direction of the edge, they could put an
# incorrectly combine the local basis functions to give the wrong global
# basis function.
#
# This issue and the use of permutations and transformations to correct
# it is discussed in detail in `Construction of arbitrary order finite
# element degree-of-freedom maps on polygonal and polyhedral cell
# meshes (Scroggs, Dokken, Richardons, Wells,
# 2021) <https://arxiv.org/abs/2102.11901>`_.
#
# Functions to permute and transform high degree elements are
# provided by Basix. In this demo, we show how these can be used.
#
# First, we import Basix and Numpy.

import basix
import numpy as np
from basix import ElementFamily, CellType, LagrangeVariant, LatticeType

# Degree 5 Lagrange element
# =========================
#
# We create a degree 5 Lagrange element on a triangle, then print the
# values of the attributes `dof_transformations_are_identity` and
# `dof_transformations_are_permutations`.
#
# The value of `dof_transformations_are_identity` is False: this tells
# us that permutations or transformations are needed for this element.
#
# The value of `dof_transformations_are_permutations` is True: this
# tells us that for this element, all the corrections we need to apply
# permutations. This is the simpler case, and means we make the
# orientation corrections by applying permutations when creating the
# DOF map.

lagrange = basix.create_element(
    ElementFamily.P, CellType.triangle, 5, LagrangeVariant.equispaced)
print(lagrange.dof_transformations_are_identity)
print(lagrange.dof_transformations_are_permutations)

# We can apply permutations by using the matrices returned by the
# method `base_transformations`. This method will return one matrix
# for each edge of the cell (for 2D and 3D cells), and two matrices
# for each face of the cell (for 3D cells). These describe the effect
# of reversing the edge or reflecting and rotating the face.
#
# For this element, we know that the base transformations will be
# permutation matrices.

print(lagrange.base_transformations())

# The matrices returned by `base_transformations` are quite large, and
# are equal to the identity matrix except for a small block of the
# matrix. It is often easier and more efficient to use the matrices
# returned by the method `entity_transformations` instead.
#
# `entity_transformations` returns a dictionary that maps the type
# of entity (`"interval"`, `"triangle"`, `"quadrilateral"`) to a
# matrix describing the effect of permuting that entity on the DOFs
# on that entity.
#
# For this element, we see that this method returns one matrix for
# an interval: this matrix reverses the order of the four DOFs
# associated with that edge.

print(lagrange.entity_transformations())

# In orders to work out which DOFs are associated with each edge,
# we use the attribute `entity_dofs`. For example, the following can
# be used to see which DOF numbers are associated with edge (dim 1)
# number 2:

print(lagrange.entity_dofs[1][2])

# Degree 2 Lagrange element
# =========================
#
# For a degree 2 Lagrange element, no permutations or transformations
# are needed. We can verify this by checking that
# `dof_transformations_are_identity` is `True`. To confirm that the
# transformations are identity matrices, we also print the base
# transformations.

lagrange_degree_2 = basix.create_element(
    ElementFamily.P, CellType.triangle, 2, LagrangeVariant.equispaced)
print(lagrange_degree_2.dof_transformations_are_identity)
print(lagrange_degree_2.base_transformations())

# Degree 2 Nédélec element
# ========================
#
# For a degree 2 Nédélec (first kind) element on a tetrahedron, the
# corrections are not all permutations, so both
# `dof_transformations_are_identity` and
# `dof_transformations_are_permutations` are `False`.

nedelec = basix.create_element(ElementFamily.N1E, CellType.tetrahedron, 2)
print(nedelec.dof_transformations_are_identity)
print(nedelec.dof_transformations_are_permutations)

# For this element, `entity_transformations` returns a dictionary
# with two entries: a matrix for an interval that describes
# the effect of reversing the edge; and an array of two matrices
# for a triangle. The first matrix for the triangle describes
# the effect of rotating the triangle. The second matrix describes
# the effect of reflecting the triangle.
#
# For this element, the matrix describing the effect of rotating
# the triangle is
#
# .. math::
#    \left(\begin{array}{cc}-1&-1\\1&0\end{array}\right).
#
# This is not a permutation, so this must be applied when assembling
# a form and cannot be applied to the DOF numbering in the DOF map.

print(nedelec.entity_transformations())

# To demonstrate how these transformations can be used, we create a
# lattice of points where we will tabulate the element.

points = basix.create_lattice(
    CellType.tetrahedron, 5, LatticeType.equispaced, True)

# If (for example) the direction of edge 2 in the physical cell does
# not match its direction on the reference, then we need to adjust the
# tabulated data.
#
# As the cell sub-entity that we are correcting is an interval, we
# get the `"interval"` item from the entity transformations dictionary.
# We use `entity_dofs[1][2]` (1 is the dimension of an edge, 2 is the
# index of the edge we are reversing) to find out which dofs are on
# our edge.
#
# To adjust the tabulated data, we loop over each point in the lattice
# and over the value size. For each of these values, we apply the
# transformation matrix to the relevant DOFs.

data = nedelec.tabulate(0, points)

transformation = nedelec.entity_transformations()["interval"][0]
dofs = nedelec.entity_dofs[1][2]

for point in range(data.shape[1]):
    for dim in range(data.shape[3]):
        data[0, point, dofs, dim] = np.dot(transformation, data[0, point, dofs, dim])

print(data)

# More efficient functions that apply the transformations and
# permutations directly to data can be used via Basix's C++
# interface.
#
# C++ demo
# ========
# The following C++ code runs the same demo using Basix's C++ interface:
#
# .. literalinclude:: ../cpp/demo_dof_transformations/main.cpp
#    :language: c++



# Copyright (c) 2021 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import pytest
import numpy as np

cells = ["triangle", "quadrilateral", "tetrahedron", "hexahedron", "pyramid", "prism"]


@pytest.mark.parametrize("cell", cells)
def test_volume(cell):
    cell_type = getattr(basix.CellType, cell)
    volumes = {
        "triangle": 1 / 2,
        "quadrilateral": 1,
        "tetrahedron": 1 / 6,
        "hexahedron": 1,
        "pyramid": 1 / 3,
        "prism": 1 / 2
    }
    assert np.isclose(basix.cell.volume(cell_type), volumes[cell])


@pytest.mark.parametrize("cell", cells)
def test_normals(cell):
    cell_type = getattr(basix.CellType, cell)
    normals = basix.cell.facet_normals(cell_type)
    facets = basix.topology(cell_type)[-2]
    geometry = basix.geometry(cell_type)

    for normal, facet in zip(normals, facets):
        assert np.isclose(np.linalg.norm(normal), 1)
        for v in facet[1:]:
            tangent = geometry[v] - geometry[facet[0]]
            assert np.isclose(np.dot(tangent, normal), 0)


@pytest.mark.parametrize("cell", cells)
def test_outward_normals(cell):
    cell_type = getattr(basix.CellType, cell)
    normals = basix.cell.facet_outward_normals(cell_type)
    facets = basix.topology(cell_type)[-2]
    geometry = basix.geometry(cell_type)
    midpoint = sum(geometry) / len(geometry)

    for normal, facet in zip(normals, facets):
        assert np.dot(geometry[facet[0]] - midpoint, normal) > 0


@pytest.mark.parametrize("cell", cells)
def test_facet_orientations(cell):
    cell_type = getattr(basix.CellType, cell)
    normals = basix.cell.facet_normals(cell_type)
    outward_normals = basix.cell.facet_outward_normals(cell_type)
    orientations = basix.cell.facet_orientations(cell_type)

    for n1, n2, orient in zip(normals, outward_normals, orientations):
        if orient:
            assert np.allclose(n1, -n2)
        else:
            assert np.allclose(n1, n2)


@pytest.mark.parametrize("cell", cells)
def test_sub_entity_connectivity(cell):
    cell_type = getattr(basix.CellType, cell)
    connectivity = basix.cell.sub_entity_connectivity(cell_type)
    topology = basix.topology(cell_type)

    assert len(connectivity) == len(topology)

    for dim, entities in enumerate(connectivity):
        assert len(entities) == len(topology[dim])
        for n, entity in enumerate(entities):
            for dim2, connected_entities in enumerate(entity):
                for n2 in connected_entities:
                    if dim > dim2:
                        for i in topology[dim2][n2]:
                            assert i in topology[dim][n]
                    else:
                        for i in topology[dim][n]:
                            assert i in topology[dim2][n2]



import os
import pytest
import subprocess
import sys

# Get all the demos in this folder
path = os.path.dirname(os.path.realpath(__file__))
demos = []
for file in os.listdir(path):
    if file.endswith(".py") and file.startswith("demo"):
        demos.append(file)


@pytest.mark.parametrize("demo", demos)
def test_demo(demo):
    ret = subprocess.run([sys.executable, demo], cwd=str(path), check=True)
    assert ret.returncode == 0



"""Functions to manipulate quadrature types."""

from ._basixcpp import QuadratureType as _QT


def string_to_type(rule: str) -> _QT:
    """Convert a string to a Basix QuadratureType enum.

    Args:
        rule: The quadrature rule as a string.

    Returns:
        The quadrature type
    """
    if rule == "default":
        return _QT.Default

    if rule in ["Gauss-Lobatto-Legendre", "GLL"]:
        return _QT.gll
    if rule in ["Gauss-Legendre", "GL", "Gauss-Jacobi"]:
        return _QT.gauss_jacobi
    if rule == "Xiao-Gambutas":
        return _QT.xiao_gambutas

    if not hasattr(_QT, rule):
        raise ValueError(f"Unknown quadrature rule: {rule}")
    return getattr(_QT, rule)


def type_to_string(quadraturetype: _QT) -> str:
    """Convert a Basix QuadratureType enum to a string.

    Args:
        quadraturetype: The quadrature type

    Returns:
        The quadrature rule as a string.
    """
    return quadraturetype.name



# Copyright (c) 2021 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import pytest
import numpy
import sympy

one = sympy.Integer(1)
x = sympy.Symbol("x")
y = sympy.Symbol("y")
z = sympy.Symbol("z")


@pytest.mark.parametrize("degree", range(6))
@pytest.mark.parametrize("cell_type", [
    basix.CellType.interval, basix.CellType.triangle, basix.CellType.quadrilateral,
    basix.CellType.tetrahedron, basix.CellType.hexahedron,
])
def test_legendre(cell_type, degree):
    points, weights = basix.make_quadrature(cell_type, 2 * degree)

    polys = basix.tabulate_polynomials(basix.PolynomialType.legendre, cell_type, degree, points)

    matrix = numpy.empty((polys.shape[1], polys.shape[1]))
    for i, col_i in enumerate(polys.T):
        for j, col_j in enumerate(polys.T):
            matrix[i, j] = sum(col_i * col_j * weights)

    assert numpy.allclose(matrix, numpy.identity(polys.shape[1]))


def evaluate(function, pt):
    if len(pt) == 1:
        return function.subs(x, pt[0])
    elif len(pt) == 2:
        return function.subs(x, pt[0]).subs(y, pt[1])
    elif len(pt) == 3:
        return function.subs(x, pt[0]).subs(y, pt[1]).subs(z, pt[2])


# TODO: pyramid
@pytest.mark.parametrize("cell_type, functions, degree", [
    [basix.CellType.interval, [one, x], 1],
    [basix.CellType.interval, [one, x, x**2], 2],
    [basix.CellType.interval, [one, x, x**2, x**3], 3],
    [basix.CellType.triangle, [one, y, x], 1],
    [basix.CellType.triangle, [one, y, x, y**2, x * y, x**2], 2],
    [basix.CellType.triangle, [one, y, x, y**2, x * y, x**2, y**3, x*y**2, x**2*y, x**3], 3],
    [basix.CellType.tetrahedron, [one], 0],
    [basix.CellType.tetrahedron, [one, z, y, x], 1],
    [basix.CellType.tetrahedron, [one, z, y, x, z**2, y * z, x * z, y**2, x * y, x**2], 2],
    [basix.CellType.tetrahedron, [one, z, y, x, z**2, y * z, x * z, y**2, x * y, x**2,
                                  z**3, y * z**2, x * z**2, y**2 * z, x * y * z, x**2 * z,
                                  y**3, x * y**2, x**2 * y, x**3], 3],
    [basix.CellType.quadrilateral, [one, y, x, x * y], 1],
    [basix.CellType.quadrilateral, [one, y, y**2, x, x * y, x * y**2, x**2, x**2 * y, x**2 * y**2], 2],
    [basix.CellType.hexahedron, [one, z, y, y * z, x, x * z, x * y, x * y * z], 1],
    [basix.CellType.prism, [one, z, y, y * z, x, x * z], 1],
    [basix.CellType.prism, [one, z, z**2, y, y * z, y * z**2, x, x * z, x * z**2,
                            y**2, y**2 * z, y**2 * z**2, x * y, x * y * z, x * y * z**2,
                            x**2, x**2 * z, x**2 * z**2], 2],
])
def test_order(cell_type, functions, degree):
    points, weights = basix.make_quadrature(cell_type, 2 * degree)
    polys = basix.tabulate_polynomials(basix.PolynomialType.legendre, cell_type, degree, points)

    assert len(functions) == polys.shape[1]

    eval_points = basix.create_lattice(cell_type, 10, basix.LatticeType.equispaced, True)
    eval_polys = basix.tabulate_polynomials(basix.PolynomialType.legendre, cell_type, degree, eval_points)

    for n, function in enumerate(functions):
        expected_eval = [float(evaluate(function, i)) for i in eval_points]

        # Using n polynomials
        # The monomial should NOT be exactly represented using this many
        coeffs = []
        values = numpy.array([evaluate(function, i) for i in points])
        for p in range(n):
            coeffs.append(sum(values * polys[:, p] * weights))
        actual_eval = [float(sum(coeffs * p[:n])) for p in eval_polys]
        assert not numpy.allclose(expected_eval, actual_eval)

        # Using n+1 polynomials
        # The monomial should be exactly represented using this many
        coeffs = []
        values = numpy.array([evaluate(function, i) for i in points])
        for p in range(n + 1):
            coeffs.append(sum(values * polys[:, p] * weights))
        actual_eval = [float(sum(coeffs * p[:n+1])) for p in eval_polys]
        assert numpy.allclose(expected_eval, actual_eval)
